/*
 * linux/drivers/net/cs89x0_time.c : Reference No : 43PJ292-20050906-QFP_BOARD01-000
 *
 * Internal Network driver for TC90400
 *
 *  (c) Copyright TOSHIBA CORPORATION SEMICONDUCTOR COMPANY  2003-2005 
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA
 *
 */

#include <linux/config.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/types.h>
#include <linux/fcntl.h>
#include <linux/interrupt.h>
#include <linux/ptrace.h>
#include <linux/ioport.h>
#include <linux/in.h>
#include <linux/slab.h>
#include <linux/string.h>
#include <linux/init.h>
#include <asm/system.h>
#include <asm/bitops.h>
#include <asm/io.h>
#include <linux/errno.h>
#include <linux/spinlock.h>

#include <linux/netdevice.h>
#include <linux/etherdevice.h>
#include <linux/skbuff.h>

#include <asm/tc90400/tc90400.h>
#include "cs89x0.h"

/* TC90400 Timer (32-bit registers) */
#define TC90400_TMR_CHBASE(ch)		(TC90400_TMR_BASE + (ch * 0x100))
#define TC90400_TMR_TMTCR(ch)		(TC90400_TMR_CHBASE(ch) + 0x00)
#define TC90400_TMR_TMTISR(ch)          (TC90400_TMR_CHBASE(ch) + 0x04)
#define TC90400_TMR_TMCPRA(ch)          (TC90400_TMR_CHBASE(ch) + 0x08)
#define TC90400_TMR_TMCPRB(ch)          (TC90400_TMR_CHBASE(ch) + 0x0c)
#define TC90400_TMR_TMITMR(ch)          (TC90400_TMR_CHBASE(ch) + 0x10)
#define TC90400_TMR_TMCCDR(ch)          (TC90400_TMR_CHBASE(ch) + 0x20)
#define TC90400_TMR_TMPGMR(ch)          (TC90400_TMR_CHBASE(ch) + 0x30)
#define TC90400_TMR_TMTRR(ch)           (TC90400_TMR_CHBASE(ch) + 0xf0)

/* number of Timer channel */
#define TMTCR_NUM_OF_CHANNEL	4

/* Timer Count Enable */
#define TMTCR_TCE_MASK      (0x1 << 7)
#define TMTCR_TCE_HALT      (0x0 << 7)
#define TMTCR_TCE_OPERATE   (0x1 << 7)

/* Counter Clock Divide Enable */
#define TMTCR_CCDE_MASK     (0x1 << 6)
#define TMTCR_CCDE_DISABLE  (0x0 << 6)
#define TMTCR_CCDE_ENABLE   (0x1 << 6)

/* Counter Reset Enable */
#define TMTCR_CRE_MASK      (0x1 << 5)
#define TMTCR_CRE_DISABLE   (0x0 << 5)
#define TMTCR_CRE_ENABLE    (0x1 << 5)

/* External Clock Edge Select */
#define TMTCR_ECES_MASK         (0x1 << 3)
#define TMTCR_ECES_FALLING_EDGE (0x0 << 3)
#define TMTCR_ECES_RAISING_EDGE (0x1 << 3)

/* Counter Clock Select */
#define TMTCR_CCS_MASK      (0x1 << 2)
#define TMTCR_CCS_INTERNAL  (0x0 << 2)
#define TMTCR_CCS_EXTERNAL  (0x1 << 2)

/* Timer Mode */
#define TMTCR_TMODE_MASK    (0x3)
#define TMTCR_TMODE_IT      (0x0)   /* Interval Timer Mode */
#define TMTCR_TMODE_PG      (0x1)   /* Pulse Generator Mode */
#define TMTCR_TMODE_WT      (0x2)   /* Watchdog Timer Mode */

/* Timer Interval Interrupt Enable */
#define TMITMR_TIIE_MASK    (0x1 << 15)
#define TMITMR_TIIE_DISABLE (0x0 << 15)
#define TMITMR_TIIE_ENABLE  (0x1 << 15)

/* Interval Timer Zero Clear Enable */
#define TMITMR_TZCE_MASK    (0x1)
#define TMITMR_TZCE_DISABLE (0x0)
#define TMITMR_TZCE_ENABLE  (0x1)

/*
 * irq: IRQ No.
 * dev: not used
 * regs: not used
*/
void tc90400_timer_ch_intr(int irq, void * dev_id, struct pt_regs * regs)
{
	int ch;

	if( (irq > TC90400_IRQ_TMR0)
			&& (irq < (TC90400_IRQ_TMR0 + TMTCR_NUM_OF_CHANNEL)) )
		ch = irq - TC90400_IRQ_TMR0;
	else
		return;

	/* Acknowledge and reset the interrupt */
	TC90400_WR(TC90400_TMR_TMTISR(ch),0);

}

extern int tc90400_cpu_clock;

/*
 * ch: Timer channel no.(0..3)
 * dev: not used
 * interval: timeer interval (usec)
*/
void tc90400_time_ch_setup( int ch, struct net_device *dev, int interval )
{
	u32	temp;

	/* TMR Init */
	TC90400_WR(TC90400_TMR_TMTCR(ch),
			( TMTCR_CCDE_ENABLE
			| TMTCR_CRE_ENABLE
			| TMTCR_CCS_INTERNAL
			| TMTCR_TMODE_IT));

	TC90400_WR(TC90400_TMR_TMTISR(ch),0);
	TC90400_WR(TC90400_TMR_TMCCDR(ch),0);

	temp = ((tc90400_cpu_clock / 2) / 2);	/* E-BUS CLK */
	temp = (temp / 2) / 1000; /* Div2 / 1000Hz=1ms */
	temp = temp * interval / 1000;
#if 0	/* debug */
	printk("%s: tc90400_cpu_clock=%d Compare value=%d\n"
						, dev->name, tc90400_cpu_clock,temp);
#endif
	TC90400_WR(TC90400_TMR_TMCPRA(ch),temp);
}

/*
 * ch: Timer channel no.(0..3)
 * dev: not used
*/
void tc90400_time_ch_start( int ch, struct net_device *dev )
{
	u32	temp;

	/* Enable Interrupt */
        temp = (TMITMR_TIIE_ENABLE | TMITMR_TZCE_ENABLE);
        TC90400_WR(TC90400_TMR_TMITMR(ch),temp);

        /* Start Counter */
        temp = (TC90400_RD(TC90400_TMR_TMTCR(ch)) & (~TMTCR_CRE_ENABLE));
        TC90400_WR(TC90400_TMR_TMTCR(ch),temp);
        TC90400_WR(TC90400_TMR_TMTCR(ch),(temp|TMTCR_TCE_OPERATE));
}

/*
 * ch: Timer channel no.(0..3)
 * dev: not used
*/
void tc90400_time_ch_exit( int ch, struct net_device *dev )
{
        TC90400_WR(TC90400_TMR_TMTCR(ch),0);
}


